#include <iostream>
#include <fstream>
#include <cstring>
#include <cstdio>
#include <cmath>
#include "rplot.h"
extern "C" {
#include "ceil.h"
}

#define ERROR(msg,status) cerr << "Error: " << (msg) << "\n"; exit(status)
#define WARNING(msg) cerr << "Warning: " << (msg) << "\n"
#define NO_SIGNAL -99.00

enum FILE_FORMAT {
  FORMAT_POSTSCRIPT,
  FORMAT_PNG
};

void Usage(char **argv) {
  cerr << "Usage: cat file1.dat [file2.dat ...] | " << argv[0] 
       << " [options] > image.[png|eps]\n"
    "Options:\n"
    "         -date YYYYMMDD output only data written on this date\n"
    "      -averays n        average n rays\n"
    "     -avegates n        average n gates\n"
    "    -hourrange hr1 hr2  set hour axis range\n"
    "     -duration hrs      set duration of plot (overridden by `-hourrange')\n"
    "  -heightrange ht1 ht2  set height axis range\n"
    "         -size <x>x<y>  set size of figure for cloud plots\n"
    "          -png          set image format to portable network graphics (default)\n"
    "     -ps, -eps          set image format to encapsulated postscript\n"
    "          -xfs server   specify the X font server\n"
    "  -noisefactor f        set noise rejection threshold (default 4)\n"
    "      -noclean          don't clean noisy rays\n"
    "      -verbose          talk to me\n";
  exit(0);
}

int main(int argc,char **argv) {
  RFigure *figure;
  RAxes *axes, *cbar;
  RPalette colormap;
  RFontLoader *font_loader;

  FILE_FORMAT format=FORMAT_PNG;
  char datestr[32];

  LCTime time;
  LCCloudBases cloudbases;
  LCBackscatterProfile profile_hr,profile;

  unsigned short int iarg=0,r_year=0,r_month=0,r_day=0;
  int i, daymod=0;
  bool bydate=false, sethourrange=false, setheightrange=false;
  bool verbose=false, quiet=false;
  double starthour=0, endhour=24, minheight=0, maxheight=12,dhour=1.0;
  int width=800,height=300;
  int averays=1, avegates=1, status, cloudgate;
  char *xfs = "";
  int rays_read;

  float logbeta[LC_MAX_GATES], range[LC_MAX_GATES];
  bool firstray=true, plotdate=true, cleanray=true, not_finished=true;
  float decimalhour, lastdecimalhour=-1, stdn, noisefactor=4.0, duration=6.0;

  /* Process command-line arguments */

  while (++iarg < argc) {
    if (strcmp(argv[iarg],"-date") == 0) {
      bydate=true;
      if (sscanf(argv[++iarg],"%04hu%02hu%02hu",
		 &r_year,&r_month,&r_day) < 3) {
	ERROR("Failed to read date: should be in the form YYYYMMDD",1);
      }
    }
    else if (strcmp(argv[iarg],"-hourrange") == 0) {
      if (sscanf(argv[++iarg],"%lf",&starthour) < 1) {
	ERROR("Failed to read start hour as double",1);
      }
      if (sscanf(argv[++iarg],"%lf",&endhour) < 1) {
	ERROR("Failed to read start hour as double",1);
      }
      sethourrange=true;
    }
    else if (strcmp(argv[iarg],"-heightrange") == 0) {
      if (sscanf(argv[++iarg],"%lf",&minheight) < 1) {
	ERROR("Failed to read minimum height as double",1);
      }
      if (sscanf(argv[++iarg],"%lf",&maxheight) < 1) {
	ERROR("Failed to read maximum height as double",1);
      }
      setheightrange=true;
    }
    else if (strcmp(argv[iarg],"-size") == 0) {
      if (sscanf(argv[++iarg],"%dx%d",&width,&height) < 2) {
	ERROR("Failed to read size: should be in the form <width>x<height>",
	      1);
      }
    }
    else if (strcmp(argv[iarg],"-avegates") == 0 && iarg+1<argc) {
      avegates = atoi(argv[++iarg]);
      if (avegates < 1) {
	WARNING("Cannot average negative or zero gates");
	avegates = 1;
      }
    }
    else if (strcmp(argv[iarg],"-averays") == 0 && iarg+1<argc) {
      averays = atoi(argv[++iarg]);
      if (averays < 1) {
	WARNING("Cannot average negative or zero rays");
	averays = 1;
      }
    }
    else if (strcmp(argv[iarg],"-duration") == 0) {
      duration=atof(argv[++iarg]);
    }
    else if (strcmp(argv[iarg],"-noisefactor") == 0) {
      noisefactor=atof(argv[++iarg]);
    }
    else if (strcmp(argv[iarg],"-noclean") == 0) {
      cleanray=false;
    }
    else if ((strcmp(argv[iarg],"-ps") == 0) || (strcmp(argv[iarg],"-eps")
						 == 0)) {
      format = FORMAT_POSTSCRIPT;
    }
    else if (strcmp(argv[iarg],"-png") == 0) {
      format = FORMAT_PNG;
    }
    else if (strcmp(argv[iarg],"-verbose") == 0) {
      verbose=true;
    }
    else if (strcmp(argv[iarg],"-quiet") == 0) {
      quiet=true;
    }
    else if (strcmp(argv[iarg],"-xfs") == 0) {
      if (argc > iarg+1) {
	xfs = argv[++iarg];
      }
      else {
	ERROR("X font server not specified",1);
      }
    }
    else {
      cerr << "Error: `" << argv[iarg] << "' not understood\n";
      Usage(argv);
    }
  }


  if (bydate) {
    sethourrange = 1;
  }

  try {
    while (not_finished) {
      rays_read = (LCGetAveragedRay(stdin,&time,&cloudbases,&profile_hr,
				    averays));

      if (rays_read) {
	/* PROCESS DATA */
	LCAverageRay(&profile_hr,&profile,avegates);
	if (cleanray) {
	  LCCustomCleanRay(&cloudgate,&stdn,&profile,8,noisefactor);
	}
	decimalhour = ((float) time.hour) + ((float) time.minute)/60.0 +
	  ((float) time.second)/3600.0;
      }
      else {
	not_finished = false;
	break;
      }

      if (format == FORMAT_POSTSCRIPT) {
	figure = new RFigure(width,height,POSTSCRIPT, &cout,
			     "Chilbolton lidar ceilometer", 0.5);
      }
      else {
	figure = new RFigure(width,height,BITMAP_8BIT);
      }

      font_loader = new RFontLoader(xfs, figure->is_bitmap());
      font_loader->verbose(verbose);

      figure->clear();
      figure->axis(0,0,1,1);
      figure->font = font_loader->load(12,string("helvetica"),string("medium"),
				       string("r"),string("adobe"));
      figure->text(0.01,0.0,"<B>Chilbolton 905 nm CT75K Lidar Ceilometer",
		   ANCHOR_BOTTOMLEFT);

      axes = new RAxes(figure,0.07,0.13,0.89,0.9);

      create_jet_palette(&colormap,64);

      while ((status=(LCGetAveragedRay(stdin,&time,&cloudbases,&profile_hr,
				       averays)))) {

	LCAverageRay(&profile_hr,&profile,avegates);
	if (cleanray) {
	  LCCustomCleanRay(&cloudgate,&stdn ,&profile,8,noisefactor);
	}

	if (bydate && (r_year != time.year || r_month != time.month
		       || r_day != time.day)) {
	  continue;
	}

	decimalhour = ((float) time.hour) + ((float) time.minute)/60.0 +
	  ((float) time.second)/3600.0 + 24*daymod;
	plotdate=false;
	if (lastdecimalhour>decimalhour) {
	  daymod++;
	  decimalhour += 24;
	  plotdate=true;
	}
	lastdecimalhour=decimalhour;

	if ((decimalhour>endhour) && (!firstray)) {
	  not_finished = false;
	  break;
	}
	else if (decimalhour<starthour) {
	  continue;
	}

	if (plotdate) {
	  LCGetDate(datestr,&time);
	  axes->text(decimalhour,maxheight,datestr,ANCHOR_BOTTOMLEFT);
	}

	if (firstray) {
	  firstray=false;
	  for (i=0;i<profile.ngates;i++) {
	    range[i]=((float) profile.range[i]) / 1000.0;
	  }
	  if (!sethourrange) {
	    starthour=floor(decimalhour);
	    endhour=starthour+duration;
	  }
	  if (endhour <= starthour) {
	    cerr << "Error: End hour is less than start hour!\n";
	    break;
	  }
	  axes->axis(starthour,minheight,endhour,maxheight);
	  axes->clipbox_to_axis();
	  LCGetDate(datestr,&time);
	  axes->text(decimalhour,maxheight,datestr,ANCHOR_BOTTOMLEFT);
	}
	for (i=0;i<profile.ngates;i++) {
	  if (profile.beta[i] > 0) {
	    logbeta[i]=log10((double) profile.beta[i]);
	  }
	  else {
	    logbeta[i]=NO_SIGNAL;
	  }
	}
	if (verbose) {
	  cerr << ".";
	}
	/* Use `*0.01' rather than /120.0 (because of 30 second rays)
	   to avoid blank pixel swaths due to rounding errors */
	axes->vray(decimalhour, decimalhour + averays*0.01, range, logbeta,
		   profile.ngates, &colormap, 0, 3, NO_SIGNAL);
      }
      if (verbose) {
	cerr << "\n";
      }
      axes->box();
      axes->xtick(24,NULL,0,0,1);
      dhour = (endhour-starthour > 6) ? floor((endhour-starthour)/6.0) :
	(1.0+floor((endhour-starthour)/4.0))/2.0;
      axes->xticks_time(0,dhour);
      axes->yticks(0,1);
      axes->xlabel("Time (UTC)");
      axes->ylabel("Height (km)");

      cbar = new RAxes(figure,0.96,0.23,0.98,0.8);
      cbar->colorbar(&colormap,VERTICAL);
      cbar->box();
      cbar->axis(0,-7,1,-4);
      cbar->yticks(-7,1,"10<SUP>%g",0);
      cbar->ylabel("Lidar backscatter coefficient "
		   "(sr<SUP>-1</SUP> m<SUP>-1</SUP>)");

      if (format == FORMAT_PNG) {
	figure->write_png(&cout);
      }
      delete cbar;
      delete axes;
      delete figure;
      delete font_loader;
    }
  }
  catch  (const char * error_message) {
    cerr << error_message << endl;
    return 1;
  }

  exit(0);
}

